@@ -26,6 +26,7 @@
 
 #define __GPA_KEY_C__
 
+#include <config.h>
 #include <string.h>
 #include <libxml/globals.h>
 #include "gpa-utils.h"
@@ -38,13 +39,13 @@ static void gpa_key_init (GPAKey *key);
 
 static void gpa_key_finalize (GObject *object);
 
-static GPANode *gpa_key_duplicate (GPANode *node);
-static gboolean gpa_key_verify (GPANode *node);
-static guchar *gpa_key_get_value (GPANode *node);
-static gboolean gpa_key_set_value (GPANode *node, const guchar *value);
-static GPANode *gpa_key_get_child (GPANode *node, GPANode *ref);
-static GPANode *gpa_key_lookup (GPANode *node, const guchar *path);
-static void gpa_key_modified (GPANode *node, guint flags);
+static GPANode * gpa_key_duplicate (GPANode *node);
+static gboolean  gpa_key_verify    (GPANode *node);
+static guchar *  gpa_key_get_value (GPANode *node);
+static gboolean  gpa_key_set_value (GPANode *node, const guchar *value);
+static GPANode * gpa_key_lookup    (GPANode *node, const guchar *path);
+static GPANode * gpa_key_get_child (GPANode *node, GPANode *previous_child);
+static void      gpa_key_modified  (GPANode *node, guint flags);
 
 static gboolean gpa_key_merge_children_from_key (GPAKey *dst, GPAKey *src);
 
@@ -86,20 +87,20 @@ gpa_key_class_init (GPAKeyClass *klass)
 	object_class->finalize = gpa_key_finalize;
 
 	node_class->duplicate = gpa_key_duplicate;
-	node_class->verify = gpa_key_verify;
+	node_class->verify    = gpa_key_verify;
 	node_class->get_value = gpa_key_get_value;
 	node_class->set_value = gpa_key_set_value;
 	node_class->get_child = gpa_key_get_child;
-	node_class->lookup = gpa_key_lookup;
-	node_class->modified = gpa_key_modified;
+	node_class->lookup    = gpa_key_lookup;
+	node_class->modified  = gpa_key_modified;
 }
 
 static void
 gpa_key_init (GPAKey *key)
 {
 	key->children = NULL;
-	key->option = NULL;
-	key->value = NULL;
+	key->option   = NULL;
+	key->value    = NULL;
 }
 
 static void
@@ -111,20 +112,22 @@ gpa_key_finalize (GObject *object)
 
 	while (key->children) {
 		if (G_OBJECT (key->children)->ref_count > 1) {
-			g_warning ("GPAKey: Child %s has refcount %d\n", GPA_NODE_ID (key->children), G_OBJECT (key->children)->ref_count);
+			g_warning ("GPAKey: Child %s has refcount %d\n", gpa_node_id (key->children), G_OBJECT (key->children)->ref_count);
 		}
 		key->children = gpa_node_detach_unref_next (GPA_NODE (key), key->children);
 	}
 
-	if (key->option) {
-		gpa_node_unref (key->option);
-		key->option = NULL;
-	}
+	g_assert (key->option);
+	gpa_node_unref (key->option);
+	key->option = NULL;
 
+	/* FIXME: how can we have a key without a value? (Chema)
+	 * g_assert (key->value);
+	 */
 	if (key->value) {
 		g_free (key->value);
-		key->value = NULL;
 	}
+	key->value = NULL;
 
 	G_OBJECT_CLASS (parent_class)->finalize (object);
 }
@@ -134,33 +137,37 @@ gpa_key_duplicate (GPANode *node)
 {
 	GPAKey *key, *new;
 	GPANode *child;
-	GSList *l;
+	GSList *list;
 
 	key = GPA_KEY (node);
 
-	new = (GPAKey *) gpa_node_new (GPA_TYPE_KEY, GPA_NODE_ID (node));
+	new = (GPAKey *) gpa_node_new (GPA_TYPE_KEY, gpa_node_id (node));
 
-	if (key->value) new->value = g_strdup (key->value);
+	if (key->value)
+		new->value = g_strdup (key->value);
 
 	if (key->option) {
 		new->option = key->option;
 		gpa_node_ref (new->option);
 	}
 
-	l = NULL;
+	list = NULL;
 	for (child = key->children; child != NULL; child = child->next) {
 		GPANode *newchild;
 		newchild = gpa_node_duplicate (child);
-		if (newchild) l = g_slist_prepend (l, newchild);
+		if (newchild)
+			list = g_slist_prepend (list, newchild);
 	}
 
-	while (l) {
+	while (list) {
 		GPANode *newchild;
-		newchild = GPA_NODE (l->data);
-		l = g_slist_remove (l, newchild);
+		
+		newchild = GPA_NODE (list->data);
 		newchild->parent = GPA_NODE (new);
 		newchild->next = new->children;
 		new->children = newchild;
+
+		list = g_slist_remove (list, newchild);
 	}
 
 	return GPA_NODE (new);
@@ -169,7 +176,7 @@ gpa_key_duplicate (GPANode *node)
 static gboolean
 gpa_key_verify (GPANode *node)
 {
-	/* fixme: Verify on option */
+	/* FIXME: Verify on option (Lauris) */
 
 	return TRUE;
 }
@@ -187,6 +194,47 @@ gpa_key_get_value (GPANode *node)
 	return NULL;
 }
 
+static gboolean
+gpa_key_set_value (GPANode *node, const guchar *value)
+{
+	GPAOption *option;
+	GPAKey *key;
+
+	g_return_val_if_fail (node != NULL, FALSE);
+	g_return_val_if_fail (GPA_IS_KEY (node), FALSE);
+	g_return_val_if_fail (value != NULL, FALSE);
+	g_return_val_if_fail (GPA_KEY (node)->value != NULL, FALSE);
+
+	key    = GPA_KEY (node);
+	option = GPA_KEY_OPTION (key);
+	
+	if (strcmp (key->value, value) == 0)
+		return TRUE;
+
+	if (GPA_OPTION_IS_LIST (option)) {
+		GPAOption *child;
+		child = gpa_option_get_child_by_id (option, value);
+		if (child) {
+			/* Everything is legal */
+			g_free (key->value);
+			key->value = g_strdup (value);
+			gpa_key_merge_children_from_option (key, child);
+			gpa_node_unref (GPA_NODE (child));
+			gpa_node_request_modified (node, 0);
+			return TRUE;
+		}
+	} else if (GPA_OPTION_IS_STRING (option)) {
+		g_free (key->value);
+		key->value = g_strdup (value);
+		gpa_node_request_modified (node, 0);
+		return TRUE;
+	}
+
+	return FALSE;
+}
+
+/* FIXME: I Think i broke set_value. Old version : (Chema) */
+#if 0
 static gboolean
 gpa_key_set_value (GPANode *node, const guchar *value)
 {
@@ -231,38 +279,36 @@ gpa_key_set_value (GPANode *node, const guchar *value)
 
 	return FALSE;
 }
+#endif
 
 static GPANode *
-gpa_key_get_child (GPANode *node, GPANode *ref)
+gpa_key_get_child (GPANode *node, GPANode *previous_child)
 {
-	GPAKey *key;
+	GPAKey *key = GPA_KEY (node);
 
-	key = GPA_KEY (node);
-
-	if (!ref) {
-		if (key->children)
-			gpa_node_ref (key->children);
-		return key->children;
+	if (previous_child) {
+		if (previous_child->next)
+			gpa_node_ref (previous_child->next);
+		return previous_child->next;
 	}
-
-	if (ref->next)
-		gpa_node_ref (ref->next);
-
-	return ref->next;
+	
+	if (key->children)
+		gpa_node_ref (key->children);
+	
+	return key->children;
 }
 
 static GPANode *
 gpa_key_lookup (GPANode *node, const guchar *path)
 {
 	GPAKey *key;
-	GPANode *child;
+	GPANode *child = NULL;
 	const guchar *dot, *next;
 	gint len;
 
 	key = GPA_KEY (node);
 
-	child = NULL;
-	if (gpa_node_lookup_ref (&child, GPA_NODE (key->option), path, "Option"))
+	if (gpa_node_lookup_helper (GPA_NODE (key->option), path, "Option", &child))
 		return child;
 
 	dot = strchr (path, '.');
@@ -275,10 +321,11 @@ gpa_key_lookup (GPANode *node, const guchar *path)
 	}
 
 	for (child = key->children; child != NULL; child = child->next) {
-		const guchar *cid;
+		const guchar *id;
+		
 		g_assert (GPA_IS_KEY (child));
-		cid = GPA_NODE_ID (child);
-		if (cid && strlen (cid) == len && !strncmp (cid, path, len)) {
+		id = gpa_node_id (child);
+		if (id && strlen (id) == len && !strncmp (id, path, len)) {
 			if (!next) {
 				gpa_node_ref (child);
 				return child;
@@ -303,7 +350,7 @@ gpa_key_modified (GPANode *node, guint flags)
 	while (child) {
 		GPANode *next;
 		next = child->next;
-		if (GPA_NODE_FLAGS (child) & GPA_MODIFIED_FLAG) {
+		if (GPA_NODE_FLAGS (child) & GPA_NODE_MODIFIED_FLAG) {
 			gpa_node_ref (child);
 			gpa_node_emit_modified (child, 0);
 			gpa_node_unref (child);
@@ -341,8 +388,9 @@ gpa_key_write (xmlDocPtr doc, GPANode *node)
 
 	if (option->type != GPA_OPTION_TYPE_KEY) {
 		xmln = xmlNewDocNode (doc, NULL, "Key", NULL);
-		xmlSetProp (xmln, "Id", GPA_NODE_ID (node));
-		if (key->value) xmlSetProp (xmln, "Value", key->value);
+		xmlSetProp (xmln, "Id", gpa_node_id (node));
+		if (key->value)
+			xmlSetProp (xmln, "Value", key->value);
 
 		for (child = key->children; child != NULL; child = child->next) {
 			xmlc = gpa_key_write (doc, child);
@@ -359,40 +407,46 @@ gpa_key_write (xmlDocPtr doc, GPANode *node)
 gboolean
 gpa_key_merge_from_tree (GPANode *key, xmlNodePtr tree)
 {
-	xmlChar *xmlid, *xmlval;
-	xmlNodePtr xmlc;
+	xmlNodePtr xml_node;
+	xmlChar *key_value;
+	xmlChar *key_id;
 
 	g_return_val_if_fail (key != NULL, FALSE);
 	g_return_val_if_fail (GPA_IS_KEY (key), FALSE);
 	g_return_val_if_fail (tree != NULL, FALSE);
-	g_return_val_if_fail (!strcmp (tree->name, "Key"), FALSE);
+	g_return_val_if_fail (strcmp (tree->name, "Key") == 0, FALSE);
 
-	xmlid = xmlGetProp (tree, "Id");
-	g_assert (xmlid);
-	g_assert (GPA_NODE_ID_COMPARE (key, xmlid));
-	xmlFree (xmlid);
+	key_id = xmlGetProp (tree, "Id");
+	g_assert (key_id);
+	g_assert (GPA_NODE_ID_COMPARE (key, key_id));
+	xmlFree (key_id);
 
-	xmlval = xmlGetProp (tree, "Value");
-	if (xmlval) {
-		gpa_node_set_value (key, xmlval);
-		xmlFree (xmlval);
+	key_value = xmlGetProp (tree, "Value");
+	if (key_value) {
+		gpa_node_set_value (key, key_value);
+		xmlFree (key_value);
 	}
 
-	for (xmlc = tree->xmlChildrenNode; xmlc != NULL; xmlc = xmlc->next) {
-		if (!strcmp (xmlc->name, "Key")) {
-			xmlChar *keyid;
-			keyid = xmlGetProp (xmlc, "Id");
-			if (keyid) {
-				GPANode *kch;
-				for (kch = GPA_KEY (key)->children; kch != NULL; kch = kch->next) {
-					if (GPA_NODE_ID_COMPARE (kch, keyid)) {
-						gpa_key_merge_from_tree (kch, xmlc);
-						break;
-					}
-				}
-				xmlFree (keyid);
+	for (xml_node = tree->xmlChildrenNode; xml_node != NULL; xml_node = xml_node->next) {
+		if (strcmp (xml_node->name, "Key") == 0) {
+			xmlChar *key_id;
+			GPANode *child_key;
+			
+			key_id = xmlGetProp (xml_node, "Id");
+			if (!key_id || !*key_id) {
+				g_warning ("Invalid or missing key id while loading a GPAKey [%s]\n",
+					   gpa_node_id (key));
+				continue;
 			}
-		}
+			for (child_key = GPA_KEY (key)->children; child_key != NULL; child_key = child_key->next) {
+				if (GPA_NODE_ID_COMPARE (child_key, key_id)) {
+					gpa_key_merge_from_tree (child_key, xml_node);
+					break;
+				}
+			}
+			
+			xmlFree (key_id);
+		} 
 	}
 
 	return TRUE;
