@@ -28,7 +28,9 @@
 
 #define __GP_TRANSPORT_LPR_C__
 
+#include "config.h"
 #include <libgnomeprint/gnome-print.h>
+
 #include "gp-transport-lpr.h"
 
 static void gp_transport_lpr_class_init (GPTransportLPRClass *klass);
@@ -143,13 +145,14 @@ gp_transport_lpr_open (GnomePrintTransport *transport)
 	}
 
 	tlpr->pipe = popen (command, "w");
-	g_free (command);
 
 	if (tlpr->pipe == NULL) {
-		g_warning ("Opening 'lpr -P%s' for output failed", tlpr->printer);
+		g_warning ("Opening '%s' for output failed", command);
+		g_free (command);
 		return GNOME_PRINT_ERROR_UNKNOWN;
 	}
-
+	
+	g_free (command);
 	return GNOME_PRINT_OK;
 }
 
